@@ -25,9 +25,11 @@
 # any later version.
 
 import sys
+import inspect
 from ..overrides import override
 from ..importer import modules
 
+
 if sys.version_info >= (3, 0):
     _basestring = str
     _callable = lambda c: hasattr(c, '__call__')
@@ -99,12 +101,27 @@ class Pad(Gst.Pad):
     def __init__(self):
         self._real_chain_func = None
         self._real_event_func = None
+        self._real_query_func = None
 
     def _chain_override(self, pad, parent, buf):
         return self._real_chain_func(pad, buf)
 
-    def _event_override(self, pad, parent, buf):
-        return self._real_event_func(pad, buf)
+    def _event_override(self, pad, parent, event):
+        return self._real_event_func(pad, event)
+
+    def _query_override(self, pad, parent, query):
+        query.mini_object.refcount -= 1
+        try:
+            res = self._real_query_func(pad, query)
+        except TypeError:
+            try:
+                res = self._real_query_func(pad, parent, query)
+            except TypeError:
+                raise TypeError("Invalid query method %s, 2 or 3 arguments requiered"
+                                % self._real_query_func)
+        query.mini_object.refcount += 1
+
+        return res
 
     def set_chain_function(self, func):
         self._real_chain_func = func
@@ -114,6 +131,14 @@ class Pad(Gst.Pad):
         self._real_event_func = func
         self.set_event_function_full(self._event_override, None)
 
+    def set_query_function(self, func):
+        self._real_query_func = func
+        self.set_query_function_full(self._chain_override, None)
+
+    def set_query_function_full(self, func, udata):
+        self._real_query_func = func
+        self._real_set_query_function_full(self._query_override, None)
+
     def query_caps(self, filter=None):
         return Gst.Pad.query_caps(self, filter)
 
@@ -123,6 +148,7 @@ class Pad(Gst.Pad):
             raise LinkError(ret)
         return ret
 
+Pad._real_set_query_function_full = Gst.Pad.set_query_function_full
 Pad = override(Pad)
 __all__.append('Pad')
 
@@ -131,7 +157,7 @@ class GhostPad(Gst.GhostPad):
         if direction is None:
             if target is None:
                 raise TypeError('you must pass at least one of target'
-                        'and direction')
+                                'and direction')
             direction = target.props.direction
 
         Gst.GhostPad.__init__(self, name=name, direction=direction)
@@ -199,6 +225,7 @@ __all__.append('Pipeline')
 class Structure(Gst.Structure):
     def __getitem__(self, key):
         return self.get_value(key)
+
     def __setitem__(self, key, value):
         return self.set_value(key, value)
 
@@ -309,3 +336,60 @@ Gst.warning = _gi_gst.warning
 Gst.error = _gi_gst.error
 Gst.fixme = _gi_gst.fixme
 Gst.memdump = _gi_gst.memdump
+
+# Make sure PyGst is not usable if GStreamer has not been initialized
+class NotInitialized(Exception):
+    pass
+__all__.append('NotInitialized')
+
+def fake_method(*args):
+    raise NotInitialized("Please call Gst.init(argv) before using GStreamer")
+
+
+real_functions = [o for o in inspect.getmembers(Gst) if isinstance(o[1], type(Gst.init))]
+
+class_methods = []
+for cname_klass in [o for o in inspect.getmembers(Gst) if isinstance(o[1], type(Gst.Element)) or isinstance(o[1], type(Gst.Caps))]:
+    class_methods.append((cname_klass,
+                         [(o, cname_klass[1].__dict__[o])
+                          for o in cname_klass[1].__dict__
+                          if isinstance(cname_klass[1].__dict__[o], type(Gst.init))]))
+
+def init_pygst():
+    for fname, function in real_functions:
+        if fname not in ["init", "init_check", "deinit"]:
+            setattr(Gst, fname, function)
+
+    for cname_class, methods in class_methods:
+        for mname, method in methods:
+            setattr(cname_class[1], mname, method)
+
+
+def deinit_pygst():
+    for fname, func in real_functions:
+        if fname not in ["init", "init_check", "deinit"]:
+            setattr(Gst, fname, fake_method)
+    for cname_class, methods in class_methods:
+        for mname, method in methods:
+            setattr(cname_class[1], mname, fake_method)
+
+real_init = Gst.init
+def init(argv):
+    init_pygst()
+    return real_init(argv)
+Gst.init = init
+
+real_init_check = Gst.init_check
+def init_check(argv):
+    init_pygst()
+    return real_init_check(argv)
+Gst.init_check = init_check
+
+real_deinit = Gst.deinit
+def deinit():
+    deinit_pygst()
+    return real_deinit()
+
+Gst.deinit = deinit
+
+deinit_pygst()
